(* $Id: godi_plist.mli 749 2008-05-22 23:38:27Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2004 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(** Parsing and printing of PLIST files *)


type core_plist_element =
    [ `Dirrm of string
    | `File of string * string option   (* file name, checksum *)
    | `Comment of string
    | `Exec of string
    | `Unexec of string
    | `Option of [ `Extract_in_place | `Preserve ]
    | `Ignore
    | `Unsupported of string * string
	 (* @mode, @owner, @group, @mtree, @ignore_inst *)
    ]
    (** The subset of [inst_plist_element] that actually refers to files *)

type inst_plist_element =
    [ `Name of string    (* name plus full version string *)
    | `Cwd of string
    | `Src of string     (* FIXME: Questionable *)
    | `Pkgdep of Godi_versions.dependency
    | `Blddep of Godi_versions.dependency
    | `Pkgcfl of Godi_versions.dependency  (* Note: not understood in pkg db *)
    | `Display of string
    | core_plist_element
    ]
  (** An [inst_plist_element] represents a line in the +CONTENTS file, i.e.
   * the variant of PLIST used in binary packages.
   *)

type user_plist_element =
    [ `Findlib    of user_plist_option * string
    | `Dllfindlib of user_plist_option * string
    | `Flatdir    of user_plist_option * string
    | `Deepdir    of user_plist_option * string
    | `Glob       of user_plist_option * string
    | `Library    of user_plist_option * string * string
    | `Stubdll    of user_plist_option * string * string
    | `File       of user_plist_option * string  (* no checksum *)
    | `Comment    of string
    | `Exec       of string
    | `Unexec     of string
    | `Option     of [ `Extract_in_place | `Preserve ]
    | `Dirrm      of string
    | `Unsupported of string * string
    ]
  (** A [user_plist_element] represents a line in the PLIST file written
    * by the packager.
   *)

and user_plist_option =
    { upo_optional : bool;      (** @optional keyword *)
      upo_ocamlopt : upo_bool;  (** @ocamlopt keyword *)
      upo_ocamloptp : upo_bool; (** @ocamloptp keyword *)
      upo_ocamldll : upo_bool;  (** @ocamldll keyword *)
    }
  (** The conditions for which the PLIST element holds *)

and upo_bool = [ `None | `IfAvail | `IfNotAvail ]
  (** [`None] means no condition, [`IfAvail] means to interpret the element
    * only if the feature is availanle, and [`IfNotAvail] means to interpret
    * only if the feature is not available.
   *)

exception Plist_error of string * string
  (** The first string is the file name, the second string the errorneous
   * line. This exception is raised when a known directive is used wrong.
   *)

exception Plist_processing_error of string * string
  (** (file, error text) *)


val parse_inst_plist : ?no_files:bool -> string -> inst_plist_element list
  (** Parses the file, and returns the list of directives contained in it.
   * It is ensured that there is a `Cwd before the first `File, `Exec and
   * `Unexec element. Furthermore, the PLIST is normalized:
   * - @ignore directives are applied, causing that the following directives
   *   are removed
   * - @cwd directives are normalized, such that there is only one @cwd
   *   left.
   *
   * @param no_files If set to true, only [`Name], [`Cwd], [`Src],
   * [`Pkgdep], [`Blddep], [`Pkgcfl], [`Option] directives are returned.
   *)

val parse_inst_plist_string : ?no_files:bool -> string -> inst_plist_element list
  (** Parses the PLIST from the string value *)

(*
val apply_ignore_directive : inst_plist_element list -> inst_plist_element list
  (** Applies the `Ignore directive, and returns the transformed list (i.e.
    * ignored elements are removed)
   *)
 *)

(*
val apply_cwd_directive : inst_plist_element list -> inst_plist_element list
  (** Removes all `Cwd directives but the first. The `File directives are
   * changed accordingly. The placeholders in `Exec and `Unexec are
   * partially resolved (i.e. %D is replaced by %D/<relpath>).
   *)
 *)

val extract_package_name : inst_plist_element list -> string
  (** Extracts the package name. Raises Not_found if not found.
    * The name includes the full version string.
  *)

val extract_prefix : inst_plist_element list -> string
  (** Returns the prefix of the package, i.e. the first `Cwd directive *)

val extract_pkgdep : inst_plist_element list -> Godi_versions.dependency list
val extract_blddep : inst_plist_element list -> Godi_versions.dependency list
val extract_pkgcfl : inst_plist_element list -> Godi_versions.dependency list
  (** Return the @pkgdep, @blddep, and @pkgcfl elements, respectively. *)

val extract_display : inst_plist_element list -> string
  (** Returns the @display or Not_found *)

val print_inst_plist : 
       Netchannels.out_obj_channel -> inst_plist_element list -> unit
  (** Print the inst_plist to the channel *)

(*
val prepare_for_pkgdb : inst_plist_element list -> inst_plist_element list
 *) (* THis is nonsense *)
  (** Transforms the PLIST so it is suitable for the +CONTENTS file of
    * the package db. In particular, these transformations are applied:
    * - [`Pkgcfl] elements are removed
   *)

val parse_user_plist : string -> user_plist_element list
  (** Parses the file, and returns the list of directives contained in it.
   *)


exception Plist_expansion_error of string
  (** Expansion error *)

val expand_plist : root:string -> 
                   ocamlopt_feature:bool ->
                   ocamloptp_feature:bool ->
                   ocamldll_feature:bool ->
                   library_re:(string -> Str.regexp) ->
                   stubdll_re:(string -> Str.regexp) ->
                   cygwin_fixups:bool ->
                   user_plist_element list -> 
                     core_plist_element list
  (** Expands the user elements and returns core elements. The [root]
    * argument is the root of the file hierarchy where files are looked
    * up. The bool arguments specify whether features are available.
    *
    * library_re: Returns the regexp used to match library file names
    *   (@library) for a given library name.
    *
    * stubdll_re: Returns the regexp used to match DLL file names
    *   (@stubdll) for a given DLL name.
    *
    * The returned [`Files] directives have checksums (unless they are
    * symlinks).
   *)
