(* $Id: godi_types.ml 749 2008-05-22 23:38:27Z gerd $
 * ----------------------------------------------------------------------
 * Copyright 2003 by Gerd Stolpmann.
 *
 * This file is part of godi-tools.
 *
 * godi-tools is free software; you can redistribute it and/or modify it
 * under the conditions in the file LICENSE.
 *
 * godi-tools is distributed WITHOUT ANY WARRANTY, as detailed in the
 * file LICENSE.
 *)

(* Common types for GODI tools *)

open Godi_versions
open Godi_exec
open Godi_plist
open Printf

exception User_error of string


let user_error msg = raise(User_error msg);;


let exec_to_user_error ?(pref_user_error=false) f arg =
  try
    f arg
  with
      Command_failure(cmd,args,code) when pref_user_error ->
        ( match code with
              WEXITED n ->
                user_error(sprintf "Command fails with code %d: %s" n cmd)
            | WSIGNALED n ->
                user_error(sprintf "Command terminated because of signal %d: %s" n cmd)
            | WSTOPPED _ ->
                user_error(sprintf "Command stopped: %s" cmd)
        )
    | Command_not_found cmd when pref_user_error ->
        user_error(sprintf "Command not found: %s" cmd)
    (* Note: Command_discontinued is very special, and not turned into
     * a User_error
     *)
 ;;


exception Restart_recommended of string;;
exception Restart of string;;
exception File_handler_error of string

type console =
    { protocol : collector;
      interactivity : Godi_exec.interactivity
    }

let stderr_console() =
  { protocol = `Forward (fun s -> prerr_string s; flush stderr);
    interactivity = { Godi_exec.period = (-1.0); callback = fun() -> true }
  }

let printmsg console msg =
  Godi_exec.collect console.protocol msg

class type file_handler =
object
  method url : string
  method list : console -> unit -> string list
  method get : console -> string -> (string * (unit->unit)option)
end


type user_status =
    [ `Keep
    | `Build
    | `Remove
    | `Add
    | `Build_dep of int
    | `Remove_dep of int
    | `Add_dep of int
    ]
;;


class var_id = object end;;

let new_var_id () = new var_id;;


class type any_package =
object
  method name : string
  method version : string
  method pkgrevision : int

  method comment : string
  method description : string
  method godi_maintainer : string

  method dependencies : dependency list
  method build_dependencies : dependency list
  method conflicts : dependency list

  method user_status : user_status
  method set_user_status : user_status -> unit

  method var_id : var_id

end


class type mixin_variant =
object('self)
  method variant : var_id -> 'self
end


class type source_package =
object
  inherit any_package
  inherit mixin_variant

  method source_directory : string
  method build_message : string
  method install_message : string
  method build_options : string list
  method values_of_build_options : (string * string) list
  method values_of_user_build_options : (string * string) list
  method set_user_build_options : (string * string) list -> unit
  method config_dialog_shown : bool
  method set_config_dialog_shown : bool -> unit
  method self_update_pkg : string
  method rescan_deps : console -> string -> var_id -> unit
end


class type binary_package =
object
  inherit any_package
  inherit mixin_variant

  method binary_archive : string
  method require_binary_archive : console -> unit
  method release_binary_archive : unit -> unit

end


class type installed_package =
object
  inherit any_package
  inherit mixin_variant

  method db_directory : string
  method install_message : string
  method build_version : string
  method build_info : string
  method preserve : bool
  method inst_plist : inst_plist_element list
  method size : int64
  method size_all : int64
  method requirements_script_path : string option
  method install_script_path : string option
  method deinstall_script_path : string option
  method requirements_script : string option
  method install_script : string option
  method deinstall_script : string option
  method db_entry_last_modified : float
  method update_successor_file : ?filter:(string -> bool) -> console -> unit
end

class type dispersion =
object
  method root : string option
  method files : string list
  method size : int64
  method plist : Godi_plist.core_plist_element list
end


class type dispersed_package =
object
  inherit any_package
  inherit mixin_variant
  method dispersion : dispersion 
  method meta_data : (string * string) list
  method install_message : string
  method build_version : string
  method build_info : string
  method preserve : bool
  method inst_plist : inst_plist_element list
  method size : int64
  method size_all : int64
  method requirements_script : string option
  method install_script : string option
  method deinstall_script : string option
end


type package =
    [ `Source_pkg of source_package
    | `Binary_pkg of binary_package
    | `Installed_pkg of installed_package
    | `Dispersed_pkg of dispersed_package
    ]
;;


type pkgtype =
    [ `Source
    | `Binary
    | `Installed
    | `Dispersed
    ]
;;


let source_pkg =
  function
      `Source_pkg p -> p
    | _ -> failwith "Godi_types.source_pkg: package has bad type "
;;


let binary_pkg =
  function
      `Binary_pkg p -> p
    | _ -> failwith "Godi_types.binary_pkg: package has bad type "
;;


let installed_pkg =
  function
      `Installed_pkg p -> p
    | _ -> failwith "Godi_types.installed_pkg: package has bad type "
;;


let dispersed_pkg =
  function
      `Dispersed_pkg p -> p
    | _ -> failwith "Godi_types.dispersed_pkg: package has bad type "
;;


let any_pkg =
  function
      `Source_pkg p -> (p :> any_package)
    | `Binary_pkg p -> (p :> any_package)
    | `Installed_pkg p -> (p :> any_package)
    | `Dispersed_pkg p -> (p :> any_package)
;;


let pkgtype =
  function
      `Source_pkg _ -> `Source
    | `Binary_pkg _ -> `Binary
    | `Installed_pkg _ -> `Installed
    | `Dispersed_pkg _ -> `Dispersed
;;


let pkg_name p = (any_pkg p)#name;;
let pkg_version p = (any_pkg p)#version;;
let pkg_pkgrevision p = (any_pkg p)#pkgrevision;;
let pkg_comment p = (any_pkg p)#comment;;
let pkg_description p = (any_pkg p)#description;;
let pkg_godi_maintainer p = (any_pkg p)#godi_maintainer;;
let pkg_dependencies p = (any_pkg p)#dependencies;;
let pkg_build_dependencies p = (any_pkg p)#build_dependencies;;
let pkg_conflicts p = (any_pkg p)#conflicts;;
let pkg_user_status p = (any_pkg p)#user_status;;

let pkg_fullname p =
  let n = pkg_name p in
  let v = pkg_version p in
  let r = pkg_pkgrevision p in
  match r with
      0 -> n ^ "-" ^ v
    | _ -> n ^ "-" ^ v ^ "godi" ^ string_of_int r
;;


let to_be_built pkg =
  match (any_pkg pkg)#user_status with
      `Build | `Build_dep _ ->
	true
    | _ ->
	false
;;


let to_be_removed pkg =
  match (any_pkg pkg) # user_status with
      `Remove | `Remove_dep _ ->
	true
    | _ ->
	false
;;


let to_be_added pkg =
  match (any_pkg pkg) # user_status with
      `Add | `Add_dep _ ->
	true
    | _ ->
	false
;;


module Pkg = struct
  type t = package
  let compare = (Pervasives.compare : t -> t -> int)
end;;


module Int = struct
  (* This module is still missing in the stdlib! *)
  type t = int
  let compare a b = a - b
end;;


module PType = struct
  type t = pkgtype
  let weight =
    function
	`Source -> 0
      | `Binary -> 1
      | `Installed -> 2
      | `Dispersed -> 3
  let compare a b =
    (weight a) - (weight b)
end;;


module PkgSet = Set.Make(Pkg);;
module StrMap = Map.Make(String);;
module IntMap = Map.Make(Int);;
module PTypeMap = Map.Make(PType);;
module PMap = Map.Make(Pkg);;


module PkgMap = struct
  type key = package
  type 'a t =
      { index_name        : PkgSet.t StrMap.t;
	index_version     : PkgSet.t StrMap.t;
	index_rev         : PkgSet.t IntMap.t;
	index_pkgtype     : PkgSet.t PTypeMap.t;
	mapping           : 'a PMap.t;
      }

  let empty =
    { index_name = StrMap.empty;
      index_version = StrMap.empty;
      index_rev = IntMap.empty;
      index_pkgtype = PTypeMap.empty;
      mapping = PMap.empty
    }

  let is_empty map = (map.mapping = PMap.empty)

  let _or_empty1 f arg1 =
    try f arg1 with Not_found -> PkgSet.empty

  let _or_empty2 f arg1 arg2 =
    try f arg1 arg2 with Not_found -> PkgSet.empty

  let add k v map =
    if PMap.mem k map.mapping then (
      (* Simple case: k is already entered into the index maps, so we only
       * have to update "mapping". Note that name, version, etc. are
       * immutable.
       *)
      { map with mapping = PMap.add k v map.mapping }
    )
    else (
      (* Complicated case: Add k to all index maps, and to "mapping": *)

      let name = pkg_name k in
      let version = pkg_version k in
      let rev = pkg_pkgrevision k in
      let pkgtype = pkgtype k in

      let cur_set_name =
	_or_empty2 StrMap.find name map.index_name in
      let cur_set_version =
	_or_empty2 StrMap.find version map.index_version in
      let cur_set_rev =
	_or_empty2 IntMap.find rev map.index_rev in
      let cur_set_pkgtype =
	_or_empty2 PTypeMap.find pkgtype map.index_pkgtype in

      let new_set_name    = PkgSet.add k cur_set_name in
      let new_set_version = PkgSet.add k cur_set_version in
      let new_set_rev     = PkgSet.add k cur_set_rev in
      let new_set_pkgtype = PkgSet.add k cur_set_pkgtype in

      { index_name    = StrMap.add name new_set_name map.index_name;
	index_version = StrMap.add version new_set_version map.index_version;
	index_rev     = IntMap.add rev new_set_rev map.index_rev;
	index_pkgtype = PTypeMap.add pkgtype new_set_pkgtype map.index_pkgtype;
	mapping       = PMap.add k v map.mapping
      }
    )

  let remove k map =
    if PMap.mem k map.mapping then (
      (* Complicated case: Remove k from all index maps, and from "mapping": *)

      let name = pkg_name k in
      let version = pkg_version k in
      let rev = pkg_pkgrevision k in
      let pkgtype = pkgtype k in

      let (cur_set_name, cur_set_version, cur_set_rev, cur_set_pkgtype) =
	try
	  (StrMap.find name map.index_name,
	   StrMap.find version map.index_version,
	   IntMap.find rev map.index_rev,
	   PTypeMap.find pkgtype map.index_pkgtype)
	with
	    Not_found -> assert false
	      (* This should not happen, and is an indicator of an illegal
	       * index.
	       *)
      in

      let new_set_name    = PkgSet.remove k cur_set_name in
      let new_set_version = PkgSet.remove k cur_set_version in
      let new_set_rev     = PkgSet.remove k cur_set_rev in
      let new_set_pkgtype = PkgSet.remove k cur_set_pkgtype in

      let update new_set index_key index remove_fn add_fn =
	if new_set = PkgSet.empty then
	  remove_fn index_key index
	else
	  add_fn index_key new_set index
      in

      { index_name =
	  update new_set_name name map.index_name StrMap.remove StrMap.add;
	index_version =
	  update new_set_version version map.index_version StrMap.remove
	    StrMap.add;
	index_rev =
	  update new_set_rev rev map.index_rev IntMap.remove IntMap.add;
	index_pkgtype =
	  update new_set_pkgtype pkgtype map.index_pkgtype PTypeMap.remove
	    PTypeMap.add;
	mapping = PMap.remove k map.mapping
      }
    )
    else
      map

  let make l =
    List.fold_left (fun acc (k,d) -> add k d acc) empty l

  let find k map =
    PMap.find k map.mapping

  let list map =
    PMap.fold (fun k d acc -> (k,d) :: acc) map.mapping []

  let keys map =
    PMap.fold (fun k d acc -> k :: acc) map.mapping []

  let values map =
    PMap.fold (fun k d acc -> d :: acc) map.mapping []


  exception Return of bool

  let for_all f map =
    try
      PMap.iter
	(fun k d ->
	   if not (f k d) then raise(Return false))
	map.mapping;
      true
    with
	Return b -> b

  let exists f map =
    try
      PMap.iter
	(fun k d ->
	   if f k d then raise(Return true))
	map.mapping;
      false
    with
	Return b -> b

  let filter_list f map =
    PMap.fold
      (fun k d acc ->
	 if f k d then (k,d) :: acc else acc)
      map.mapping
      []

  let filter f map =
    make(filter_list f map)

  let partition_list f map =
    PMap.fold
      (fun k d (acc1,acc2) ->
	 if f k d then ((k,d) :: acc1, acc2) else (acc1, (k,d) :: acc2))
      map.mapping
      ([], [])

  let partition f map =
    let (l1,l2) = partition_list f map in
    (make l1, make l2)

  let iter f map =
    PMap.iter f map.mapping

  let map f map =
    { map with mapping = PMap.map f map.mapping }

  let mapi f map =
    { map with mapping = PMap.mapi f map.mapping }

  let fold f map acc =
    PMap.fold f map.mapping acc

  let mem k map =
    PMap.mem k map.mapping

  let union map1 map2 =
    (* CHECK: Maybe this is faster if the index maps are united independently,
     * because we can use PkgSet.union for this.
     *)
    fold add map1 map2

  let diff map1 map2 =
    (* CHECK: Maybe this is faster if the index maps are diff'd independently,
     * because we can use PkgSet.diff for this.
     *)
    fold (fun k v acc -> remove k acc) map2 map1

  let inter map1 map2 =
    (* CHECK: Maybe this is faster if the index maps are inter'd independently,
     * because we can use PkgSet.inter for this.
     *)
    fold
      (fun k v acc ->
	 if mem k map2 then
	   add k v acc
	 else
	   acc)
      map1
      empty

  let find_by ?name ?version ?pkgrevision ?pkgtype map =
    let filter_by crit_opt find_fn set_opt =
      match crit_opt with
	  None -> set_opt
	| Some n ->
	    ( match set_opt with
		  None     -> Some(_or_empty1 find_fn n)
		| Some set -> Some(PkgSet.inter
		                     set
		                     (_or_empty1 find_fn n))
	    ) in
    let set_opt =
      filter_by pkgrevision (fun n -> IntMap.find n map.index_rev)
	(filter_by version (fun n -> StrMap.find n map.index_version)
	   (filter_by pkgtype (fun n -> PTypeMap.find n map.index_pkgtype)
	      (filter_by name (fun n -> StrMap.find n map.index_name) None)
	   )
	) in
    (* Note: This takes into account that Set.inter a b is faster than
     * Set.inter b a if a is a small set.
     *)
    match set_opt with
	None ->
	  list map
      | Some set ->
	  filter_list
	    (fun pkg _ -> PkgSet.mem pkg set)
	    map


  let mem_by ?name ?version ?pkgrevision ?pkgtype map =
    let lookup_by crit_opt mem_fn result =
      match crit_opt with
	  None -> result
	| Some n ->
	    result && (mem_fn n)
    in
    lookup_by pkgrevision (fun n -> IntMap.mem n map.index_rev)
      (lookup_by version (fun n -> StrMap.mem n map.index_version)
	 (lookup_by pkgtype (fun n -> PTypeMap.mem n map.index_pkgtype)
	    (lookup_by name (fun n -> StrMap.mem n map.index_name) true)
	 )
      )
end;;


let newer_package p1 p2 =
  let v1 = (any_pkg p1) # version in
  let r1 = (any_pkg p1) # pkgrevision in
  let v2 = (any_pkg p2) # version in
  let r2 = (any_pkg p2) # pkgrevision in
  ( match Godi_versions.cmp_version_string v1 v2 with
	0 ->
	  r1 > r2
      | n when n > 0 ->
	  true
      | _ ->
	  false
  )
;;


let correlate (pl : package list) =
  let rec accumulate pl =
    (* PRE: pl is sorted *)
    match pl with
	[] -> []
      | (p :: pl') ->
	  let acc = accumulate pl' in
	  ( match acc with
		[] ->
		  [ pkg_name p, [p] ]
	      | (n, l)::acc' ->
		  if pkg_name p = n then
		    (n, p::l)::acc'
		  else
		    (pkg_name p, [p]) :: acc
	  )
  in

  let cmp_pkg p1 p2 =
    compare (pkg_name p1) (pkg_name p2) in

  let pl_s = List.sort cmp_pkg pl in
  let pl_acc = accumulate pl_s in

  List.map
    (fun (n, l) ->
       let (l_src, l') = List.partition (fun p -> pkgtype p = `Source) l in
       let (l_bin, l_inst) = List.partition (fun p -> pkgtype p = `Binary) l in
       (n, l_src, l_bin, l_inst)
    )
    pl_acc
;;
